//////////////////////////////////////////////////////////////////
// Single Author Info
//   dehuber  Daniel E. Huber
//
// Group Info
//   dehuber  Daniel E. Huber
//   jjwinega Joel J. Winegarden
//////////////////////////////////////////////////////////////////

#include "image.h"

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <jpeglib.h>


RGBPixel RGBPixel::xorWith(const RGBPixel& pixel)
{
   RGBPixel xored;
   xored.r = r ^ pixel.r;
   xored.g = g ^ pixel.g;
   xored.b = b ^ pixel.b;
   return xored;
}

HSVPixel RGBPixel::toHSVPixel()
{
   HSVPixel hsvpixel;

   float fr = r/255.0, fb = b/255.0, fg = g/255.0;
   float fh, fv, fs;
   
   float min, max;

   min = fr;
   if( fg < min )
      min = fg;
   if( fb < min )
      min = fb;

   max = fr;
   if( fg > max )
      max = fg;
   if( fb > max )
      max = fb;

   fv = max;

   fs = (max != 0) ? ((max - min) / max) : 0;

   if( fs == 0 )
      fh = 0;
   else
   {
      float delta = max - min;

      if( fr == max )
         fh = (fg - fb) / delta;
      else if( fg == max )
         fh = 2.0 + (fb - fr) / delta;
      else
         fh = 4.0 + (fr - fg) / delta;

      fh = fh * 60.0;

      if( fh < 0.0 )
         fh = fh + 360.0;
   }

   hsvpixel.h = (uchar)(fh / 360.0 * 255.0);
   hsvpixel.s = (uchar)(fs * 255.0);
   hsvpixel.v = (uchar)(fv * 255.0);

   return hsvpixel;
}

RGBPixel RGBPixel::add(const RGBPixel& pixel, float alpha)
{
   RGBPixel p;
   p.r = (uchar)( (1-alpha) * r + alpha * pixel.r );
   p.g = (uchar)( (1-alpha) * g + alpha * pixel.g );
   p.b = (uchar)( (1-alpha) * b + alpha * pixel.b );

   return p;
}

void RGBPixel::setRGB( uchar _r, uchar _g, uchar _b )
{
   r = _r;
   g = _g;
   b = _b;
}

Image::Image()
{
   m_pixels = NULL;
   m_w = 0;
   m_h = 0;
}

Image::~Image()
{
   if( m_pixels != NULL )
      delete[] m_pixels;
}

void Image::create(uint w, uint h)
{
   m_pixels = new RGBPixel [w * h];
   m_w = w;
   m_h = h;
}

void Image::copyToImage( Image& img )
{
   if( img.width() != m_w || img.height() != m_h )
      return;

   memcpy( img.m_pixels, m_pixels, m_w * m_h );
}

RGBPixel Image::pixel( int row, int col )
{
   return m_pixels[ row * m_w + col ];
}



void Image::setPixel( int row, int col, const RGBPixel& pixel )
{
   memcpy( &m_pixels[row * m_w + col], &pixel, sizeof(RGBPixel) );
}


bool Image::loadFromJPEG( const char *filename )
{
   if( m_pixels != NULL )
      delete[] m_pixels;

   jpeg_decompress_struct  cinfo;
   jpeg_error_mgr          jerr;

   cinfo.err = jpeg_std_error(&jerr);

   jpeg_create_decompress(&cinfo);

   FILE* in;

   in = fopen(filename, "rb");
   if( in == NULL)
      return false;

   jpeg_stdio_src( &cinfo, in );

   jpeg_read_header( &cinfo, TRUE );

   jpeg_start_decompress( &cinfo );

   create( cinfo.output_width, cinfo.output_height );

   int bytes_per_pix = sizeof(RGBPixel);

   int row_stride = m_w * bytes_per_pix;

   JSAMPARRAY buffer;
   buffer = (*cinfo.mem->alloc_sarray)((j_common_ptr) &cinfo, JPOOL_IMAGE,
                                       row_stride, 1);

   int row = 0;
   while( cinfo.output_scanline < cinfo.output_height )
   {
      jpeg_read_scanlines( &cinfo, buffer, 1 );

      int idx = row * m_w;

      if( cinfo.output_components == bytes_per_pix )
      {
         memcpy( &m_pixels[idx], &buffer[0][0], row_stride );
      }
      else
      {
         for( uint col=0; col<m_w; col++ )
         {
            memcpy( &m_pixels[ idx ],
                    &buffer[ 0 ][ col * cinfo.output_components ],
                    cinfo.output_components );

            idx++;
         }
      }

      row++;
   }
   
   jpeg_finish_decompress( &cinfo );

   fclose( in );

   jpeg_destroy_decompress( &cinfo );

   return true;
}


void Image::xorWithImage( Image* src, Image* dst )
{
   if( m_w != src->width() || m_w != dst->width() ||
       m_h != src->height() || m_h != dst->height() )
   {
      return;
   }

   RGBPixel* dst_buffer = dst->scanLine( 0 );
   RGBPixel* src_buffer = src->scanLine( 0 );

   for( uint ind = 0; ind < m_h * m_w; ind++ )
      dst_buffer[ind] = src_buffer[ind].xorWith( m_pixels[ind] );
}

