//////////////////////////////////////////////////////////////////
// Single Author Info
//   dehuber  Daniel E. Huber
//
// Group Info
//   dehuber  Daniel E. Huber
//   jjwinega Joel J. Winegarden
//////////////////////////////////////////////////////////////////

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <memory.h>

// linux/videodev.h header must be placed after the above headers, otherwise
// an error occurs during compilation (v4lcamera.h includes videodev.h)
#include "v4lcamera.h"
#include "quickcam.h"

V4LCamera::V4LCamera()
{
   device_fd = 0;
}

V4LCamera::~V4LCamera()
{
   closeDevice();
}

bool V4LCamera::openDevice(const char* device_name)
{
   device_fd = open( device_name, O_RDWR );

   if( device_fd <= 0 )
      return false;

   m_device_name = device_name;

   getInfo( NULL );
   
   return true;
}

void V4LCamera::closeDevice()
{
   if( isOpen() )
      close( device_fd );
   device_fd = 0;
   m_device_name = "";
}

Image* V4LCamera::createImage()
{
   Image* img = new Image();
   img->create( vidwin.width, vidwin.height );
   return img;
}


void V4LCamera::getInfo( CamInfo* caminfo )
{
   if( !isOpen() )
      return;

   ioctl(device_fd, VIDIOCGCAP, &vidcap);
   ioctl(device_fd, VIDIOCGWIN, &vidwin);
   ioctl(device_fd, VIDIOCGPICT, &vidpic);
    
   vidwin.clips = vidclips;
   vidwin.clipcount = 0;

   if( caminfo != NULL )
   {
      caminfo->name         = vidcap.name;
      caminfo->type         = vidcap.type;
      caminfo->min_width    = vidcap.minwidth;
      caminfo->max_width    = vidcap.maxwidth;
      caminfo->min_height   = vidcap.minheight;
      caminfo->max_height   = vidcap.maxheight;

      caminfo->x            = vidwin.x;
      caminfo->y            = vidwin.y;
      caminfo->width        = vidwin.width;
      caminfo->height       = vidwin.height;

      caminfo->depth        = vidpic.depth;

      if( vidcap.type & VID_TYPE_MONOCHROME )
         caminfo->color = false;
      else
         caminfo->color = true;

      caminfo->device_name = m_device_name;
   }
}

void V4LCamera::getImgProps( ImgProps* imgprops )
{
   int adaptive;

   ioctl( device_fd, VIDIOCQCGADAPTIVE, &adaptive );
   imgprops->adaptive = (adaptive == 1) ? true : false;


   ioctl( device_fd, VIDIOCGPICT, &vidpic );

   imgprops->brightness = vidpic.brightness;
   imgprops->hue = vidpic.hue;
   imgprops->color = vidpic.colour;
   imgprops->contrast = vidpic.contrast;
   imgprops->depth = vidpic.depth;

   switch( vidpic.palette )
   {
      case VIDEO_PALETTE_GREY:
         imgprops->palette = PALETTE_GREY;
         break;
      case VIDEO_PALETTE_HI240:
         imgprops->palette = PALETTE_HI240;
         break;
      case VIDEO_PALETTE_RGB565:
         imgprops->palette = PALETTE_RGB565;
         break;
      case VIDEO_PALETTE_RGB555:
         imgprops->palette = PALETTE_RGB555;
         break;
      case VIDEO_PALETTE_RGB24:
         imgprops->palette = PALETTE_RGB24;
         break;
      case VIDEO_PALETTE_RGB32:
         imgprops->palette = PALETTE_RGB32;
         break;
      case VIDEO_PALETTE_YUV422:
         imgprops->palette = PALETTE_YUV422;
         break;
      case VIDEO_PALETTE_YUYV:
         imgprops->palette = PALETTE_YUYV;
         break;
      case VIDEO_PALETTE_UYVY:
         imgprops->palette = PALETTE_UYVY;
         break;
      case VIDEO_PALETTE_YUV420:
         imgprops->palette = PALETTE_YUV420;
         break;
      case VIDEO_PALETTE_YUV411:
         imgprops->palette = PALETTE_YUV411;
         break;
      case VIDEO_PALETTE_RAW:
         imgprops->palette = PALETTE_RAW;
         break;
      case VIDEO_PALETTE_YUV422P:
         imgprops->palette = PALETTE_YUV422P;
         break;
      case VIDEO_PALETTE_YUV411P:
         imgprops->palette = PALETTE_YUV411P;
         break;
   }
}


void V4LCamera::setImgProps( ImgProps* imgprops )
{
   vidpic.brightness = imgprops->brightness;
   vidpic.hue = imgprops->hue;
   vidpic.colour = imgprops->color;
   vidpic.contrast = imgprops->contrast;

   ioctl( device_fd, VIDIOCSPICT, &vidpic );

   ioctl( device_fd, VIDIOCGPICT, &vidpic );
}


/*
  
void V4LCamera::printCameraInfo()
{

   printf("    *** Camera Info ***\n");
   printf("Name:           %s\n", vidcap.name);
   printf("Type:           %d\n", vidcap.type);
   printf("Minimum Width:  %d\n", vidcap.minwidth);
   printf("Maximum Width:  %d\n", vidcap.maxwidth);
   printf("Minimum Height: %d\n", vidcap.minheight);
   printf("Maximum Height: %d\n", vidcap.maxheight);
   printf("X:              %d\n", vidwin.x);
   printf("Y:              %d\n", vidwin.y);
   printf("Width:          %d\n", vidwin.width);
   printf("Height:         %d\n", vidwin.height);
   printf("Depth:          %d\n", vidpic.depth);

   if(vidcap.type & VID_TYPE_MONOCHROME)
      printf("Color           false\n");
   else
      printf("Color           true\n");	
}
*/


void V4LCamera::prepareCapture()
{
   struct video_mbuf vidbuf;

   ioctl( device_fd, VIDIOCSWIN, &vidwin );
   
   ioctl( device_fd, VIDIOCGMBUF, &vidbuf );

   m_pixels_mmap = (uchar*) mmap( NULL,
                                  vidbuf.size,
                                  PROT_READ,
                                  MAP_SHARED,
                                  device_fd,
                                  0 );

}

int V4LCamera::captureToImage( Image* img )
{
   if( img->width() != vidwin.width || img->height() != vidwin.height )
      return V4LCamera::INVALID_DIMS;

   struct video_mmap vm;

   int framenum = 0;

   prepareCapture();
   
   vm.format = VIDEO_PALETTE_RGB24;
   vm.frame  = framenum;
   vm.width  = vidwin.width;
   vm.height = vidwin.height;

   ioctl(device_fd, VIDIOCMCAPTURE, &vm);
   ioctl(device_fd, VIDIOCSYNC, &framenum);

   int num_pixels = vidwin.width * vidwin.height;

   //RGBPixel* scan_line = img->scanLine( 0 );

   //memcpy( &scan_line[0], &m_pixels_mmap[0], num_pixels );

   RGBPixel pixel;
   for( int pixel_ind = 0; pixel_ind < num_pixels; pixel_ind++ )
   {
      pixel.r = m_pixels_mmap[ pixel_ind * 3 + 2 ];
      pixel.g = m_pixels_mmap[ pixel_ind * 3 + 1 ];
      pixel.b = m_pixels_mmap[ pixel_ind * 3  ];

      img->setPixel( pixel_ind, pixel );
   }
   
   /*
   for( int r=0; r<vidwin.height; r++ )
   {
      for( int c=0; c<vidwin.width; c++ )
      {
         printf("%d\t%d\t%d\n",
                scan_line[r * c * 3 + c*3],
                scan_line[r * c * 3 + c*3+1],
                scan_line[r * c * 3 + c*3+2]);
      }
   }
   */

   return V4LCamera::OK;
}


